package Calculator;

use Indicators qw(get_indic_info);
use Exporter;
use MLDBM qw(DB_File Storable);
use DB_File;
use Carp;
use GTUtils qw(:date);

## This object can calculate all indics and is a big cache for results
## It can store all his results in a file (and can read it after).

sub new {
  my $self = shift;
  my $class = ref($self) || $self;
  my $db = shift ||  ## hash_ref to the db
           croak "Bad arguments to $class->new\n";
  my $node = { 'db' => $db, 'indics' => {}, 'sub' => {}, 
               'dates' => [], 'rev_dates' => {} };
  bless $node, $class;
  $node->prepare();
  
  return $node;
}

# 
# Create the list of available dates
#
sub prepare {
  my $self = shift;
  my $nbre = keys %{$self->{'db'}};
  my $i = 0;
  
  while(defined($_ = each %{$self->{'db'}})) {
   if (defined($self->{'db'}{$_})) {
     $self->{'dates'}[$i] = $_;
     $self->{'rev_dates'}{$_} = $i;
     $i++;
   }
  }
  $self->{'max_dates'} = --$i;
}

#
# Retrieve the value of an indicator for a specific date
#
sub get {
  my $self = shift;
  my $indic = shift || carp "Indicateur manquant";
  my $date = shift || carp "Date manquante";

  ## Look in the cache before
  if (exists $self->{'indics'}{$indic}
      && exists $self->{'indics'}{$indic}{$date}) {
     return $self->{'indics'}{$indic}{$date};
  }

  ## Otherwise calculate it
  return $self->calculate($indic, $date, @_);
}

#
# The core of the indicator system : how they are calculated and how they
# work
#
sub calculate {
  my $self = shift;
  my $indic = shift || carp "Indicateur manquant";
  my $date = shift || carp "Date manquante";
  my ($first, $last) = @{get_indic_info($indic, 'range')};
  my ($func, $res);

  # Vrifions dj si on a les donnes du jour demand...
  # Do we have the data for the specified dates ?
  return undef if (! exists $self->{'rev_dates'}{$date});
  my $index = $self->{'rev_dates'}{$date};
  $first += $index;
  $last += $index;
  # Je vrifie si j'ai assez de donnes pour rpondre  la demande
  # Do we have enough data before and after the specific date ?
  return undef if ($first < 0 or $last > $self->{'max_dates'});
  # Je vrifie s'il ne manque pas trop de donnes au milieu...
  # Je l'enlve car cela ne cause que des problmes alors
  # qu'il est si simple de s'en passer
#  my ($t1,$t2) = (time_of_date($self->{'dates'}[$first]),
#                  time_of_date($self->{'dates'}[$last] ));
#  return undef if ( (5 / 7 / 86400 * ($t2 - $t1 + 86400)) > 
#                    (($last - $first + 1)*16/14) );
		  
  # Fabriquons / rcuprons la sub de calcul de l'indic
  # Creation (search) of the anonymous sub for the calculation
  if (exists $self->{'sub'}{$indic} and defined $self->{'sub'}{$indic}) {
    $func = $self->{'sub'}{$indic};
  } else {
    my $code = ${get_indic_info($indic, 'code')};
    $code = q#
      $func = sub {
    # . $code . q#
      }
    #;
    eval $code;
    if ($@) {
      carp "Impossible de compiler l'indicateur : $@\n";
      return undef;
    }
    $self->{'sub'}{$indic} = $func;
  }
  
  # Prparation des donnes pour la sub de calcul
  # Preparation of the data available to the calculation subroutine
  local *date = \@{$self->{'dates'}};
  local *data = \%{$self->{'db'}};
  local *ind = \$index;
  local $calc = $self;
  
  # Calcul effectif : priorit aux arguments fournis par l'indicateur
  # Real calculation
  $res = &$func(@{get_indic_info($indic, 'param')}, @_);
                                  
  # Je garde le rsultat si j'ai le droit...
  # Cache the result if allowed
  $self->{'indics'}{$indic}{$date} = $res if 
                                       (! get_indic_info($indic, 'nocache')); 

  return $res;
}

#
# Store the cache in a file
# TODO: incremental storage
#
sub store {
  my $self = shift;
  my $file = shift;
  my %data = ();
  
  tie %data, 'MLDBM', $file, O_CREAT|O_RDWR, 0640 or 
    croak "Cannot tie $file to MLDBM : $!\n";

  %data = %{$self->{'indics'}};
  
  untie %data;
  
  return 1;
}

#
# Load the cache from a file
#
sub load {
  my $self = shift;
  my $file = shift;
  my %data = ();

  tie %data, 'MLDBM', $file, O_RDONLY, 0640 or
    croak "Cannot tie $file to MLDBM : $!\n";

  %{$self->{'indics'}} = %data;

  untie %data;
  
  return 1;
}

#
# Get the date with an index
# Useful for finding the first date with dates(0)
#
sub dates {
  my ($self, $index) = @_;
  return $self->{'dates'}[$index] if $index < @{$self->{'dates'}};
  return undef;
}

#
# Clear the cache
# 
sub clear_cache {
  my $self = shift;
  %{$self->{'indics'}} = ();
}

1;
